@@ -1,4 +1,4 @@
-__version__ = '0.0.5'
+__version__ = '0.1.1'
 
 import sys
 import requests
@@ -7,18 +7,23 @@ import logging
 import time
 import json
 import busrt
-import pandas as pd
+import pyarrow as pa
+import pyarrow.csv as csv
+import pyarrow.compute as pc
 
 from datetime import datetime
 from typing import Union
-from io import StringIO, BytesIO
+from io import BytesIO
 
 from evaics.client import HttpClient
 from evaics.sdk import OID, FunctionFailed, pack, unpack
+from collections import OrderedDict
 
 from busrt.rpc import Rpc as BusRpcClient
 
-BUS_BULK_SIZE = 10_000
+MIME_CVS = 'text/csv'
+MIME_ARROW_STREAM = 'application/vnd.apache.arrow.stream'
+MIME_ARROW_FILE = 'application/vnd.apache.arrow.file'
 
 
 class HistoryDF:
@@ -47,7 +52,7 @@ class HistoryDF:
         self.client = client
         self.mlkit = None
         self.params = {'fill': '1S'}
-        self.oid_map = {}
+        self.oid_map = []
         if params_csv is not None:
             self.read_params_csv(params_csv)
 
@@ -71,8 +76,8 @@ class HistoryDF:
         Args:
             f: file path or buffer
         """
-        params = pd.read_csv(f)
-        for _, row in params.iterrows():
+        params = csv.read_csv(f).to_pylist()
+        for row in params:
             self.oid(row['oid'],
                      status=row.get('status'),
                      value=row.get('value'),
@@ -112,7 +117,8 @@ class HistoryDF:
             m['database'] = database
         if xopts is not None:
             m['xopts'] = xopts
-        self.oid_map[oid] = m
+        m['oid'] = str(oid)
+        self.oid_map.append(m)
         return self
 
     def t_start(self, t_start: Union[float, str, datetime]):
@@ -173,7 +179,7 @@ class HistoryDF:
         """
         if not self._is_mlkit_enabled():
             self.params['xopts'] = xopts
-        for _, v in self.oid_map.items():
+        for v in self.oid_map:
             v['xopts'] = xopts
         return self
 
@@ -186,16 +192,10 @@ class HistoryDF:
         """
         if not self._is_mlkit_enabled():
             self.params['database'] = database
-        for _, v in self.oid_map.items():
+        for v in self.oid_map:
             v['database'] = database
         return self
 
-    def _prepare_df_time(self, df, tz=None):
-        if tz is None:
-            tz = time.tzname[0]
-        df['time'] = pd.to_datetime(df['time'], unit='s')
-        df['time'] = df['time'].dt.tz_localize('UTC').dt.tz_convert(tz)
-
     def _get_mlkit(self):
         if self.mlkit is None:
             try:
@@ -219,20 +219,20 @@ class HistoryDF:
                 ml_url = ml_url[:-1]
         return ml_url
 
-    def _post(self, *args, accept_gzip=False, **kwargs):
+    def _post(self, *args, accept_gzip=False, headers=None, **kwargs):
         if not self.client.token:
             self.client.authenticate()
             refreshed = True
         else:
             refreshed = False
         while True:
-            headers = {
+            req_headers = {
                 'x-auth-key': self.client.token,
             }
-            if accept_gzip:
-                headers['accept-encoding'] = 'gzip'
+            if headers is not None:
+                req_headers.update(headers)
             req = requests.post(*args,
-                                headers=headers,
+                                headers=req_headers,
                                 timeout=self.client.timeout,
                                 **kwargs)
             if req.status_code == 403 and req.text.endswith('(AUTH)'):
@@ -263,15 +263,20 @@ class HistoryDF:
                 raise RuntimeError('mlkit server not specified')
             params = {'database': database, 'oid_map': self.oid_map}
             files = {}
-            if hasattr(data, 'seek'):
-                files['file'] = ('', data, 'text/csv')
-            elif isinstance(data, str):
-                files['file'] = ('', open(data, 'r'), 'text/csv')
-            elif isinstance(data, pd.DataFrame):
-                buf = StringIO()
-                data.to_csv(buf)
+            if isinstance(data, str):
+                if data.endswith('.arrows'):
+                    mime = MIME_ARROW_STREAM
+                elif data.endswith('.arrow'):
+                    mime = MIME_ARROW_FILE
+                else:
+                    mime = MIME_CVS
+                files['file'] = ('', open(data, 'rb'), mime)
+            elif isinstance(data, pa.Table):
+                buf = BytesIO()
+                with pa.ipc.new_stream(buf, data.schema) as writer:
+                    writer.write(data)
                 buf.seek(0)
-                files['file'] = ('', buf, 'text/csv')
+                files['file'] = ('', buf, MIME_ARROW_STREAM)
             else:
                 raise ValueError('unsupported data kind')
             req = self._post(f'{ml_url}/ml/api/upload.item.state_history',
@@ -291,99 +296,189 @@ class HistoryDF:
             params['time_format'] = 'raw'
         return params
 
-    def fetch(self, t_col: str = 'keep', tz: str = None):
+    def _fetch_bus(self, t_col, strict_col_order):
+        params = self._prepare_mlkit_params(t_col)
+        res = self.client.call(
+            self._get_mlkit(),
+            busrt.rpc.Request('Citem.state_history',
+                              pack(params))).wait_completed()
+        cpl = res.get_payload()
+        names = []
+        data = []
+        while True:
+            res = self.client.call(self._get_mlkit(),
+                                   busrt.rpc.Request('N',
+                                                     cpl)).wait_completed()
+            payload = res.get_payload()
+            if payload:
+                buf = BytesIO(payload)
+                while True:
+                    try:
+                        f = pa.ipc.open_stream(buf)
+                        names += f.schema.names
+                        for b in f:
+                            data += b.columns
+                    except pa.lib.ArrowInvalid:
+                        break
+                if strict_col_order:
+                    _reoder(self.oid_map, names, data, t_col=t_col)
+                return pa.Table.from_arrays(data, names)
+            else:
+                break
+
+    def _fetch_mlsrv_http(self, ml_url, t_col, strict_col_order):
+        params = self._prepare_mlkit_params(t_col)
+        req = self._post(f'{ml_url}/ml/api/query.item.state_history',
+                         headers={
+                             'accept': MIME_ARROW_STREAM,
+                             'accept-encoding': 'gzip'
+                         },
+                         json=params,
+                         stream=True)
+        if req.ok:
+            if req.headers.get('content-encoding') == 'gzip':
+                stream = gzip.GzipFile(fileobj=req.raw, mode='rb')
+            else:
+                stream = req.raw
+            names = []
+            data = []
+            while True:
+                try:
+                    f = pa.ipc.open_stream(stream)
+                    names += f.schema.names
+                    for b in f:
+                        data += b.columns
+                except pa.lib.ArrowInvalid:
+                    break
+            if strict_col_order:
+                _reoder(self.oid_map, names, data, t_col=t_col)
+            return pa.Table.from_arrays(data, names)
+        else:
+            raise FunctionFailed(f'{req.status_code} {req.text}')
+
+    def _fetch_hmi_http(self, t_col):
+        params = self.params.copy()
+        params['i'] = [m.pop('oid') for m in self.oid_map]
+        stats = self.client.call('item.state_history', params)
+        data = OrderedDict()
+        cols = {}
+        if t_col == 'keep':
+            data['time'] = stats.pop('t')
+        for (oid, proc) in zip(params['i'], self.oid_map):
+            if proc['status'] == True:
+                col = f'{oid}/status'
+                data[col] = stats.pop(col)
+            elif isinstance(proc['status'], str):
+                col = proc['status']
+                data[col] = stats.pop(f'{oid}/status')
+            if proc['value'] == True:
+                col = f'{oid}/value'
+                data[col] = stats.pop(col)
+            elif isinstance(proc['value'], str):
+                col = proc['value']
+                data[col] = stats.pop(f'{oid}/value')
+        return pa.Table.from_pydict(data)
+
+    def fetch(self,
+              t_col: str = 'keep',
+              tz: str = 'local',
+              output='arrow',
+              strict_col_order=False):
         """
         Fetch data
 
         Optional:
+            output: output format (arrow, pandas or polars)
             t_col: time column processing, "keep" - keep the column, "drop" -
-                   drop the time column, "index" - keep and make the data frame
-                   index
-            tz: time zone
+                drop the time column
+            tz: time zone (local, custom or None to keep time column as UNIX
+                timestamp), the default is "local"
+            strict_col_order: force strict column ordering
+
 
         Returns:
             a prepared Pandas DataFrame object
         """
-        params = self.params.copy()
+        if t_col not in ['keep', 'drop']:
+            raise RuntimeError('unsupported t_col op')
+        if output not in ['arrow', 'pandas', 'polars']:
+            raise RuntimeError('unsupported output type')
         if isinstance(self.client, BusRpcClient):
-            params = self._prepare_mlkit_params(t_col)
-            res = self.client.call(
-                self._get_mlkit(),
-                busrt.rpc.Request('Citem.state_history',
-                                  pack(params))).wait_completed()
-            cursor_id = unpack(res.get_payload(), raw=False)['u']
-            cpl = pack(dict(u=cursor_id, c=BUS_BULK_SIZE))
-            df = pd.DataFrame()
-            while True:
-                res = self.client.call(self._get_mlkit(),
-                                       busrt.rpc.Request('NB',
-                                                         cpl)).wait_completed()
-                payload = res.get_payload()
-                if payload:
-                    if len(df.columns):
-                        a = pd.read_csv(BytesIO(payload),
-                                        header=None,
-                                        names=df.columns)
-                        df = pd.concat([df, a])
-                        if len(a) < BUS_BULK_SIZE:
-                            break
-                    else:
-                        df = pd.read_csv(BytesIO(payload))
-                        if len(
-                                df
-                        ) < BUS_BULK_SIZE - 1:  # df rows (strings) + header (1 string)
-                            break
-                else:
-                    break
-            if t_col != 'drop':
-                self._prepare_df_time(df, tz)
+            from_ipc = True
+            df = self._fetch_bus(t_col, strict_col_order)
         else:
             ml_url = self._get_ml_url()
             if ml_url:
-                params = self._prepare_mlkit_params(t_col)
-                req = self._post(f'{ml_url}/ml/api/query.item.state_history',
-                                 json=params,
-                                 stream=True)
-                if req.ok:
-                    if req.headers.get('content-encoding') == 'gzip':
-                        stream = gzip.GzipFile(fileobj=req.raw, mode="rb")
-                    else:
-                        stream = req.raw
-                    df = pd.read_csv(stream)
-                    if t_col != 'drop':
-                        self._prepare_df_time(df, tz)
-                else:
-                    raise FunctionFailed(f'{req.status_code} {req.text}')
+                from_ipc = True
+                df = self._fetch_mlsrv_http(ml_url, t_col, strict_col_order)
             else:
-                params['i'] = list(self.oid_map)
-                stats = self.client.call('item.state_history', params)
-                for oid, proc in self.oid_map.items():
-                    if not proc['status']:
-                        del stats[f'{oid}/status']
-                    if not proc['value']:
-                        del stats[f'{oid}/value']
-                cols = {}
-                if t_col in ['keep', 'index']:
-                    times = stats.pop('t')
-                else:
-                    del stats['t']
-                    times = None
-                df = pd.DataFrame(data=stats)
-                if times is not None:
-                    df.insert(0, 'time', times)
-                for oid, proc in self.oid_map.items():
-                    if not isinstance(proc['status'], bool):
-                        cols[f'{oid}/status'] = proc['status']
-                    if not isinstance(proc['value'], bool):
-                        cols[f'{oid}/value'] = proc['value']
-                if cols:
-                    df = df.rename(columns=cols)
-                if t_col != 'drop':
-                    self._prepare_df_time(df, tz)
-                return df
-            if t_col == 'index':
-                df = df.set_index('time')
+                from_ipc = False
+                df = self._fetch_hmi_http(t_col)
+        return _finalize_data(df,
+                              t_col=t_col,
+                              output=output,
+                              tz=tz,
+                              from_ipc=from_ipc)
+
+
+def _finalize_data(df, t_col=None, output=None, tz=None, from_ipc=False):
+    if output == 'arrow':
         return df
+    if tz == 'local':
+        tz = time.tzname[0]
+    if output == 'pandas':
+        import pandas as pd
+        df = df.to_pandas()
+        if t_col != 'drop' and tz is not None:
+            if not from_ipc:
+                df['time'] = pd.to_datetime(df['time'], unit='s')
+            df['time'] = df['time'].dt.tz_localize('UTC').dt.tz_convert(tz)
+        return df
+    elif output == 'polars':
+        import polars as pl
+        df = pl.from_arrow(df)
+        if t_col != 'drop' and tz is not None:
+            if not from_ipc:
+                df = df.with_column(pl.from_epoch('time', unit='s'))
+            df = df.with_columns(df['time'].dt.replace_time_zone('UTC'))
+            df = df.with_columns(df['time'].dt.convert_time_zone(tz))
+        return df
+    else:
+        raise RuntimeError('output unsupported')
+
+
+def _reoder(oid_map, names, data, t_col=None):
+    col_names = []
+    if t_col == 'keep':
+        col_names.append('time')
+    for m in oid_map:
+        if m['status'] in ['true', '1', True]:
+            col_names.append(_default_col_name(m['oid'], 'status'))
+        elif m['status'] not in [None, False, 'false', '0', 'null']:
+            col_names.append(str(m['status']))
+        if m['value'] in ['true', '1', True]:
+            col_names.append(_default_col_name(m['oid'], 'value'))
+        elif m['value'] not in [None, False, 'false', '0', 'null']:
+            col_names.append(str(m['value']))
+    for i, col in enumerate(col_names):
+        try:
+            pos = names.index(col)
+            if pos != i:
+                _swap_entry(names, i, pos)
+                _swap_entry(data, i, pos)
+        except ValueError:
+            pass
+
+
+def _swap_entry(arr, pos1, pos2):
+    entry1 = arr.pop(pos1)
+    entry2 = arr.pop(pos2 - 1)
+    arr.insert(pos2 - 1, entry1)
+    arr.insert(pos1, entry2)
+
+
+def _default_col_name(oid, prop):
+    return f'{oid}/{prop}'
 
 
 def _history_df(self, params_csv=None):
